/*******************************************************************************
 *
 * FILE:        drivers/atm/ia_abr.c
 *
 * DESCRIPTION: Interphase 5575 PCI ATM Device driver.
 *              Contains the ABR utilities to fix the iphase5575 errata.
 *
 *
 * EDIT HISTORY:
 *
 * Peter Wang                   5/99    Ported to linux kernel 2.2.5
 * Interphase Corp.
 *
 ******************************************************************************/

#ifdef IA_ABR
static u16 get_desc (IADEV *dev) {
  u_short 		desc_num, desc1;
  u_short		tcq_wr,i,busy;
  unsigned long delta;
  static unsigned long timer = 0; // xx = 0;
  extern void desc_dbg(IADEV *iadev);

  /* Management of host tbl with returned descriptors to TCQ */
  busy = 0;
hosttbl:
  tcq_wr = readl(dev->seg_reg+TCQ_WR_PTR) & 0xffff;
  /* (i)chip+ ABR errata #3, #4 and #5 */
  /* if host_tcq_wr == tcq_wr  Do nothing */
  while (dev->host_tcq_wr != tcq_wr) {
      desc1 = *(u_short *)(dev->seg_ram + dev->host_tcq_wr);
      if (!desc1) ;
      else if (!dev->desc_tbl[desc1 -1]) {
//        printk(" Desc %d is set to 0 at time %ld\n", desc1 -1, jiffies);
          *(u_short *) (dev->seg_ram + dev->host_tcq_wr) = 0;
      }                                 
      else if (dev->desc_tbl[desc1 -1]) 
          dev->desc_tbl[desc1 -1] = 0;

      dev->host_tcq_wr += 2;
      if (dev->host_tcq_wr > dev->ffL.tcq_ed) 
          dev->host_tcq_wr = dev->ffL.tcq_st;
  }

  if(((jiffies - timer)>100)||((dev->ffL.tcq_rd==dev->host_tcq_wr))){      
     timer = jiffies; 
     i=0;
     while(i < TX_BUFFERS) {
        delta = jiffies - dev->desc_tbl[i];
        if ((dev->desc_tbl[i]) && (delta >= HZ)) {
#if 0 /* debug */
            printk("RECOVER run!! desc_tbl %d = %d  delta = %ld, time = %ld\n",
                  i,dev->desc_tbl[i], delta, jiffies);
            if ((xx < 4) && (inc == 0)) {
                xx++;
                desc_dbg(dev);
            }
#endif
            if (dev->ffL.tcq_rd == dev->ffL.tcq_st) 
                 dev->ffL.tcq_rd =  dev->ffL.tcq_ed;
            else 
                 dev->ffL.tcq_rd -= 2;
            *(u_short *)(dev->seg_ram + dev->ffL.tcq_rd) = i+1;
            dev->desc_tbl[i] = 0;
        }
        i++;
     } /* while */
  }

   if (dev->ffL.tcq_rd == dev->host_tcq_wr) { 
      if ( busy < 100) {
         busy++; udelay(10); 
         goto hosttbl; 
      }
      else return 0xFFFF;
   }    
  /* Get the next available descriptor number from TCQ */
  desc_num = *(u_short *)(dev->seg_ram + dev->ffL.tcq_rd);

  while (!desc_num || (dev->desc_tbl[desc_num -1])) {
  //    printk("Found one desc is 0 at time %ld desc =%d  ", jiffies,desc_num);
 	dev->ffL.tcq_rd += 2;
	if (dev->ffL.tcq_rd > dev->ffL.tcq_ed) 
		dev->ffL.tcq_rd = dev->ffL.tcq_st;
	if (dev->ffL.tcq_rd == dev->host_tcq_wr) {
           if ( busy < 100) {
              busy++; udelay(10);
              goto hosttbl;
           }
           else return 0xFFFF; 
        }
	desc_num = *(u_short *)(dev->seg_ram + dev->ffL.tcq_rd);
  }

  /* get system time */
  dev->desc_tbl[desc_num -1] = jiffies;
  return desc_num;
}

/* Fix for ABR VC lockup */
static void clear_lockup (struct atm_vcc *vcc, IADEV *dev) {
  u_char          	foundLockUp;
  vcstatus_t		*vcstatus;
  register u_int          inc;
  u_short                 *shd_tbl;
  u_short                 tempCellSlot, tempFract;
  unsigned long           flags;
  struct main_vc *abr_vc = (struct main_vc *)(dev->seg_ram+MAIN_VC_TABLE);  
  struct ext_vc *eabr_vc = (struct ext_vc *)(dev->seg_ram+EXT_VC_TABLE);
  u_int  i;

  if (vcc->qos.txtp.traffic_class == ATM_ABR) {
   	vcstatus = (vcstatus_t *) &(dev->testTable[vcc->vci].vc_status);
     	vcstatus->cnt++;

    	foundLockUp = 0;
	/* Check every 5th packet on this VC for lockup condition */
    	if( vcstatus->cnt == 0x05 ) {
	    abr_vc += vcc->vci;
	    eabr_vc += vcc->vci;

	    if( eabr_vc->last_desc ) {
 		/* extended VC table says there is a descriptor awaiting segmentation */
		if( (abr_vc->status & 0x07) == ABR_STATE /* 0x2 */ ) {
          	    /* ABR state says nothing is scheduled - conflict condition */
		    /* (i)chip+ ABR errata #4 and #5 */
		    /* Wait for 10 Micro sec */
                    inc = 0; while (inc++ < 500);
		    if ( (eabr_vc->last_desc) && ((abr_vc->status & 0x07) == ABR_STATE))  
                        /* At this time, it is determined that the first error occurred. */
			foundLockUp = 1;
		}
		else {
		    /* status (!= 2) indicates that there is something to segment */
		    tempCellSlot = abr_vc->last_cell_slot;
                    tempFract    = abr_vc->fraction;
		    if((tempCellSlot == dev->testTable[vcc->vci].lastTime)
                         && (tempFract == dev->testTable[vcc->vci].fract))
		    	/* No in-rate cell has been sent for the last 5       
                         packet additions.  Assume something is wrong. */
			/* (i)chip+ ABR errata #3 and #5 */ 	
	                 foundLockUp = 1; 		    
                    dev->testTable[vcc->vci].lastTime = tempCellSlot;   
                    dev->testTable[vcc->vci].fract = tempFract; 
	} /* status */ 	    } /* last descriptor */	 	   
vcstatus->cnt = 0;     	} /* vcstatus->cnt */
	
	/* Go through the error routine if a lockup was found */
        /* and reschedule this VC on ABR scheduling table */
        /* (i)chip+ ABR errata #3, #4 and #5 */
        if (foundLockUp) {
printk("LOCK UP found\n");
                save_flags(flags); cli();   
		writew(0xFFFD, dev->seg_reg+MODE_REG_0);
		/* Wait for 10 Micro sec */
                inc = 0; while (inc++ < 500);
 
                abr_vc->status &= 0xFFF8;
                abr_vc->status |= 0x0001;  /* state is idle */
		shd_tbl = (u_short *)(dev->seg_ram + ABR_SCHED_TABLE);

		for( i = 0; ((i < SCHEDSZ) && (shd_tbl[i])); i++ );
		if (i < SCHEDSZ)
                    shd_tbl[i] = vcc->vci;
                else
                    printk("ABR Segmentation may not continue on VC %x\n", vcc->vci);

        	writew(T_ONLINE, dev->seg_reg+MODE_REG_0);
		vcstatus->cnt = 0;
                restore_flags(flags);
	} /* foundLockUp */

  } /* if an ABR VC */


}

/*
 * ABR VC Table Entry
 */
typedef struct {
	u_short		f_vc_type;	/* VC type              */
	u_short		f_nrm;		/* Nrm			*/
	u_short		f_nrmexp;	/* Nrm Exp              */
	u_short		reserved6;	/* 			*/
	u_short		f_crm;		/* Crm			*/
	u_short		reserved10;	/* Reserved		*/
	u_short		reserved12;	/* Reserved		*/
	u_short		reserved14;	/* Reserved		*/
	u_short		last_cell_slot;	/* last_cell_slot_count	*/
	u_short		f_pcr;		/* Peak Cell Rate	*/
	u_short		fraction;	/* fraction		*/
	u_short		f_icr;		/* Initial Cell Rate	*/
	u_short		f_cdf;		/* */
	u_short		f_mcr;		/* Minimum Cell Rate	*/
	u_short		f_acr;		/* Allowed Cell Rate	*/
	u_short		f_status;	/* */
} f_vc_abr_entry;

/*
 * ABR VC Table Entry
 */
typedef struct {
        u_short         r_status_rdf;   /* status + RDF         */
        u_short         r_air;          /* AIR                  */
        u_short         reserved4[14];  /* Reserved             */
} r_vc_abr_entry;   

#define MRM 3
#define MIN(x,y)	((x) < (y)) ? (x) : (y)

/*
** Conversion of 24-bit cellrate (cells/sec) to 16-bit floating point format.
**
**  +----+----+------------------+-------------------------------+
**  |  R | NZ |  5-bit exponent  |        9-bit mantissa         |
**  +----+----+------------------+-------------------------------+
** 
**    R = reserverd (written as 0)
**    NZ = 0 if 0 cells/sec; 1 otherwise
**
**    if NZ = 1, rate = 1.mmmmmmmmm x 2^(eeeee) cells/sec
*/
static u16
cellrate_to_float(u32 cr)
{

#define	NZ 		0x4000
#define	M_BITS		9		/* Number of bits in mantissa */
#define	E_BITS		5		/* Number of bits in exponent */
#define	M_MASK		0x1ff		
#define	E_MASK		0x1f

	u16		flot;
	u32		tmp = cr & 0x00ffffff;
	int 		i   = 0;

	if (cr == 0)
		return 0;

	while (tmp != 1) {
		tmp >>= 1;
		i++;
	}

	if (i == M_BITS)
		flot = NZ | (i << M_BITS) | (cr & M_MASK);
	else if (i < M_BITS)
		flot = NZ | (i << M_BITS) | ((cr << (M_BITS - i)) & M_MASK);
	else
		flot = NZ | (i << M_BITS) | ((cr >> (i - M_BITS)) & M_MASK);

	return flot;
}

#if 0
/*
** Conversion of 16-bit floating point format to 24-bit cellrate (cells/sec).
*/
static u32
float_to_cellrate(u16 rate)
{
	u32	exp, mantissa, cps;

	if ((rate & NZ) == 0)
		return 0;

	exp = (rate >> M_BITS) & E_MASK;
	mantissa = rate & M_MASK;

	if (exp == 0)
		return 1;

	cps = (1 << M_BITS) | mantissa;

	if (exp == M_BITS)
		cps = cps;
	else if (exp > M_BITS)
		cps <<= (exp - M_BITS);
	else
		cps >>= (M_BITS - exp);

	return cps;
}

#endif 

static void init_abr_vc (srv_cls_param_t *srv_p) {
    srv_p->class_type = ATM_ABR;
    srv_p->pcr        = 0x055cb7;
    srv_p->mcr        = 0;
    srv_p->icr        = 0x055cb7;
    srv_p->tbe        = 0xffffff;
    srv_p->frtt       = 0x3a;
    srv_p->rif        = 0xf;
    srv_p->rdf        = 0xb;
    srv_p->nrm        = 0x4;
    srv_p->trm        = 0x7;
    srv_p->cdf        = 0x3;
    srv_p->adtf       = 50;
}

static int
ia_open_abr_vc(IADEV *dev, srv_cls_param_t *srv_p, 
                                                struct atm_vcc *vcc, u8 flag)
{
        // f_vc_ubr_entry	*vc;
	f_vc_abr_entry  *f_abr_vc;
	r_vc_abr_entry  *r_abr_vc;
	u32		icr;
	u8		trm, nrm, crm;
	u16		adtf, air, *ptr16;
	
	f_abr_vc = (f_vc_abr_entry *) (dev->seg_ram + MAIN_VC_TABLE); 
	f_abr_vc += vcc->vci;
        
        switch (flag) {
        case 1:
#if 0
	 	/* sanity check */
		if (srv_p->pcr == 0)
			return INVALID_PCR;

		if (srv_p->pcr > RATE155)
			srv_p->pcr = RATE155;

		if ((srv_p->mcr + dev->sum_mcr) > RATE155)
			return MCR_UNAVAILABLE;

		if (srv_p->mcr > srv_p->pcr)
			return INVALID_MCR;

		if (!(srv_p->icr))
			srv_p->icr = srv_p->pcr;

		if ((srv_p->icr < srv_p->mcr) || (srv_p->icr > srv_p->pcr))
			return INVALID_ICR;

		if ((srv_p->tbe < MIN_TBE) || (srv_p->tbe > MAX_TBE))
			return INVALID_TBE;

		if ((srv_p->frtt < MIN_FRTT) || (srv_p->frtt > MAX_FRTT))
			return INVALID_FRTT;

		if (srv_p->nrm > MAX_NRM)
			return INVALID_NRM;

		if (srv_p->trm > MAX_TRM)
			return INVALID_TRM;

		if (srv_p->adtf > MAX_ADTF)
			return INVALID_ADTF;
		else if (srv_p->adtf == 0)
			srv_p->adtf = 1;

		if (srv_p->cdf > MAX_CDF)
			return INVALID_CDF;

		if (srv_p->rif > MAX_RIF)
			return INVALID_RIF;

		if (srv_p->rdf > MAX_RDF)
			return INVALID_RDF;
#endif
		/* FFRED initialization */
		// f_abr_vc = (f_vc_abr_entry *)vc;
		memset ((caddr_t)f_abr_vc, 0, sizeof(f_vc_abr_entry));

		f_abr_vc->f_vc_type = ABR;

		nrm = 2 << srv_p->nrm;		/* (2 ** (srv_p->nrm +1)) */
						/* i.e 2**n = 2 << (n-1) */
		f_abr_vc->f_nrm = nrm << 8 | nrm;

		trm = 100000/(2 << (16 - srv_p->trm));
		if ( trm == 0) trm = 1;

		f_abr_vc->f_nrmexp = (((srv_p->nrm +1) & 0x0f) << 12) |
							(MRM << 8) | trm;
		crm = srv_p->tbe / nrm;
		if (crm == 0) crm = 1;

		f_abr_vc->f_crm = crm & 0xff;

		f_abr_vc->f_pcr = cellrate_to_float(srv_p->pcr);

		icr = MIN( srv_p->icr, (srv_p->tbe > srv_p->frtt) ?
				((srv_p->tbe/srv_p->frtt)*1000000) :
				(1000000/(srv_p->frtt/srv_p->tbe)));

		f_abr_vc->f_icr = cellrate_to_float(icr);

		adtf = (10000 * srv_p->adtf)/8192;
		if (adtf == 0) adtf = 1;
                // Original:  (7 - srv_p->cdf) << 12 | adtf & 0xfff;  
		f_abr_vc->f_cdf = ((7 - srv_p->cdf) << 12 | adtf) & 0xfff;
		f_abr_vc->f_mcr = cellrate_to_float(srv_p->mcr);
		f_abr_vc->f_acr = f_abr_vc->f_icr;
		f_abr_vc->f_status = 0x0042;
		break;
	case 0:

		/* RFRED initialization */
		ptr16 = (u_short *)(dev->reass_ram + RX_VC_TABLE); 
		*(ptr16 + vcc->vci) = NO_AAL5_PKT | REASS_ABR;

		r_abr_vc = (r_vc_abr_entry *)(dev->reass_ram + REASS_TABLE);
		r_abr_vc += vcc->vci;

		r_abr_vc->r_status_rdf = (15 - srv_p->rdf) & 0x000f;
		air = srv_p->pcr << (15 - srv_p->rif);
		if (air == 0) air = 1;

		r_abr_vc->r_air = cellrate_to_float(air);

		dev->testTable[vcc->vci].vc_status = VC_ACTIVE | VC_ABR;
		dev->sum_mcr	   += srv_p->mcr;
		dev->n_abr++;
		break;

	default:
		/* Won't happen as the check is done earlier */
		break;
	}

	return	0;
}



#endif  // IA_ABR
